/*
 * tgz_extract functions based on code within zlib library
 * No additional copyright added, KJD <jeremyd@computer.org>
 *
 *   This software is provided 'as-is', without any express or implied
 *   warranty.  In no event will the authors be held liable for any damages
 *   arising from the use of this software.
 *
 * untgz.c -- Display contents and/or extract file from
 * a gzip'd TAR file
 * written by "Pedro A. Aranda Guti\irrez" <paag@tid.es>
 * adaptation to Unix by Jean-loup Gailly <jloup@gzip.org>
 */

/*
  TODO:
    without -j there is a security issue as no checking is done to directories
    change to better support -d option, presently we just chdir there
*/


#include "untgz.h"


int tgz_extract_file(gzFile tgzFile, char *filename)
{
  char *iList[1];
  iList[0] = filename;
  return tgz_extract(tgzFile, 1, 1, iList, 0, NULL);
}


/** the rest mostly is untgz.c from zlib **/

/* Values used in typeflag field.  */

#define REGTYPE  '0'		/* regular file */
#define AREGTYPE '\0'		/* regular file */
#define LNKTYPE  '1'		/* link */
#define SYMTYPE  '2'		/* reserved */
#define CHRTYPE  '3'		/* character special */
#define BLKTYPE  '4'		/* block special */
#define DIRTYPE  '5'		/* directory */
#define FIFOTYPE '6'		/* FIFO special */
#define CONTTYPE '7'		/* reserved */

#define BLOCKSIZE 512

struct tar_header
{				/* byte offset */
  char name[100];		/*   0 */
  char mode[8];		/* 100 */
  char uid[8];		/* 108 */
  char gid[8];		/* 116 */
  char size[12];		/* 124 */
  char mtime[12];		/* 136 */
  char chksum[8];		/* 148 */
  char typeflag;		/* 156 */
  char linkname[100];	/* 157 */
  char magic[6];		/* 257 */
  char version[2];	/* 263 */
  char uname[32];		/* 265 */
  char gname[32];		/* 297 */
  char devmajor[8];	/* 329 */
  char devminor[8];	/* 337 */
  char prefix[155];	/* 345 */
				/* 500 */
};

union tar_buffer {
  char               buffer[BLOCKSIZE];
  struct tar_header  header;
};


/* help functions */

int getoct(char *p,int width)
{
  int result = 0;
  char c;
  
  while (width --)
    {
      c = *p++;
      if (c == ' ')
	continue;
      if (c == 0)
	break;
      result = result * 8 + (c - '0');
    }
  return result;
}

/* regular expression matching */

#define ISSPECIAL(c) (((c) == '*') || ((c) == '/'))

int ExprMatch(char *string,char *expr)
{
  while (1)
    {
      if (ISSPECIAL(*expr))
	{
	  if (*expr == '/')
	    {
	      if (*string != '\\' && *string != '/')
		return 0;
	      string ++; expr++;
	    }
	  else if (*expr == '*')
	    {
	      if (*expr ++ == 0)
		return 1;
	      while (*++string != *expr)
		if (*string == 0)
		  return 0;
	    }
	}
      else
	{
	  if (*string != *expr)
	    return 0;
	  if (*expr++ == 0)
	    return 1;
	  string++;
	}
    }
}


/* recursive make directory */
/* abort if you get an ENOENT errno somewhere in the middle */
/* e.g. ignore error "mkdir on existing directory" */
/* */
/* return 1 if OK */
/*        0 on error */

int makedir (char *newdir)
{
  char *buffer = strdup(newdir);
  char *p;
  int  len = strlen(buffer);
  
  if (len <= 0) {
    free(buffer);
    return 0;
  }
  if (buffer[len-1] == '/') {
    buffer[len-1] = '\0';
  }
  if (CreateDirectory(buffer, NULL) != 0)
    {
      free(buffer);
      return 1;
    }

  p = buffer+1;
  while (1)
    {
      char hold;
      
      while(*p && *p != '\\' && *p != '/')
        p++;
      hold = *p;
      *p = 0;
      //if ((mkdir(buffer, 0775) == -1) && (errno == ENOENT /* != EEXIST */))
      if (!CreateDirectory(buffer, NULL) && !((GetLastError()==ERROR_FILE_EXISTS) || (GetLastError()==ERROR_ALREADY_EXISTS)) )
      {
        // fprintf(stderr,"Unable to create directory %s\n", buffer);
        PrintMessage("Unable to create directory %s\n", buffer);
        free(buffer);
	  return 0;
      }
      if (hold == 0)
        break;
      *p++ = hold;
    }
  free(buffer);
  return 1;
}


/* returns a pointer to a static buffer
 * containing fname after removing all but
 * path_sep_cnt path separators
 * if there are less than path_sep_cnt
 * separators then all will still be there.
 */
char * stripPath(int path_sep_cnt, char *fname)
{
  static char buffer[1024];
  char *fname_use = fname + strlen(fname);
  register int i=path_sep_cnt;
  do
  {
    if ( (*fname_use == '/') || (*fname_use == '\\') ) 
	{ 
      i--;
	  if (i < 0) fname_use++;
	  else fname_use--;
    }
	else
      fname_use--;
  } while ((i >= 0) && (fname_use > fname));
  
  strcpy(buffer, fname_use);
  return buffer;
}

/* returns 1 if fname in list else return 0 
 * returns 0 if list is NULL or cnt is < 0
 */
int matchname (char *fname, int cnt, char *list[], int junkPaths)
{
  register char *t;
  int i;
  int path_sep;

  /* if nothing to compare with then return failure */
  if ((list == NULL) || (cnt <= 0))
    return 0;

  for (i = 0; i < cnt; i++)
  {
    /* get count of path components in current filelist entry */
    path_sep = 0;
    if (!junkPaths)
	{
      for(t = list[i]; *t != '\0'; t++)
        if ((*t == '/') || (*t == '\\'))
          path_sep++;
	}
    if (ExprMatch(stripPath(path_sep, fname), list[i]))
      return 1;
  }

  return 0; /* no match */
}


#ifdef __GNUC__
#define HUNDREDSECINTERVAL 116444772000000000LL
#else
#define HUNDREDSECINTERVAL 116444772000000000i64
#endif
void cnv_tar2win_time(time_t tartime, FILETIME *ftm)
{
#ifdef HAS_LIBC_CAL_FUNCS
		  FILETIME ftLocal;
		  SYSTEMTIME st;
		  struct tm localt;
 
		  localt = *localtime(&tartime);
		  
		  st.wYear = (WORD)localt.tm_year+1900;
		  st.wMonth = (WORD)localt.tm_mon+1;    /* 1 based, not 0 based */
		  st.wDayOfWeek = (WORD)localt.tm_wday;
		  st.wDay = (WORD)localt.tm_mday;
		  st.wHour = (WORD)localt.tm_hour;
		  st.wMinute = (WORD)localt.tm_min;
		  st.wSecond = (WORD)localt.tm_sec;
		  st.wMilliseconds = 0;
		  SystemTimeToFileTime(&st,&ftLocal);
		  LocalFileTimeToFileTime(&ftLocal,ftm);
#else
	// avoid casts further below
    LONGLONG *t = (LONGLONG *)ftm;

	// tartime == number of seconds since midnight Jan 1 1970 (00:00:00)
	// convert to equivalent 100 nanosecond intervals
	*t = UInt32x32To64(tartime, 10000000UL);

	// now base on 1601, add number of 100 nansecond intervals between 1601 & 1970
	*t += HUNDREDSECINTERVAL;  /* 116444736000000000i64; */
#endif
}


/* Tar file extraction
 * gzFile in, handle of input tarball opened with gzopen
 * int junkPaths, nonzero indicates to ignore stored path (don't create directories)
 * int iCnt, char *iList[], argv style list of files to extract, {0,NULL} for all
 * int xCnt, char *xList[], argv style list of files NOT to extract, {0,NULL} for none
 */
int tgz_extract(gzFile in, int junkPaths, int iCnt, char *iList[], int xCnt, char *xList[])
{
  union  tar_buffer buffer;
  int    len;
  int    err;
  int    getheader = 1;
  int    remaining = 0;
  HANDLE   outfile = INVALID_HANDLE_VALUE;
  char   fname[BLOCKSIZE];
  time_t tartime;
  
  while (1)
  {
    len = gzread(in, &buffer, BLOCKSIZE);
    if (len < 0)
    {
      PrintMessage(gzerror(in, &err));
      gzclose(in);
      return -1;
    }
    /*
     * Always expect complete blocks to process
     * the tar information.
     */
    if (len != BLOCKSIZE)
    {
      PrintMessage("gzread: incomplete block read");
      gzclose(in);
      return -1;
    }
      
    /*
     * If we have to get a tar header
     */
    if (getheader == 1)
    {
      /*
       * if we met the end of the tar
       * or the end-of-tar block,
       * we are done
       */
      if ((len == 0)  || (buffer.header.name[0]== 0)) break;

      tartime = (time_t)getoct(buffer.header.mtime,12);
      strcpy(fname,buffer.header.name);
      /* LogMessage("buffer.header.name is:");  LogMessage(fname); */
	  
      switch (buffer.header.typeflag)
      {
        case DIRTYPE:
            if (!junkPaths) makedir(fname);
	      break;
        case REGTYPE:
        case AREGTYPE:
	      remaining = getoct(buffer.header.size,12);
	      if ( (remaining) && 
               ( (iList == NULL) || (matchname(fname, iCnt, iList, junkPaths)) ) &&
               (!matchname(fname, xCnt, xList, junkPaths))
             )
	      {
	          if ((outfile = CreateFile(fname,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL)) == INVALID_HANDLE_VALUE)
	          {
	              if (!junkPaths)
	              {
	                /* try creating directory */
	                char *p = strrchr(fname, '/');
	                if (p != NULL) 
	                {
	                  *p = '\0';
	                  makedir(fname);
	                  *p = '/';
	                  outfile = CreateFile(fname,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	                }
	              }
	              else
	              {
	                /* try ignoring directory */
	                char *p = strrchr(fname, '/');
	                if (p != NULL) 
	                {
	                  // be sure terminating '\0' is copied and
	                  // use ansi memcpy equivalent that handles overlapping regions
	                  //memmove(fname, p+1, strlen(p+1) + 1 );
	                  MoveMemory(fname, p+1, strlen(p+1) + 1 );

	                  outfile = CreateFile(fname,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	                }
	              }
	          }
			  PrintMessage("%s%s", (outfile == INVALID_HANDLE_VALUE)?"Could not create file ":"Writing ", fname);
	      }
	      else
	          outfile = INVALID_HANDLE_VALUE;

	      /*
	       * could have no contents
	       */
	      getheader = (remaining) ? 0 : 1;
	      break;
        default:
/*
	      if (action == TGZ_LIST)
	          printf(" %s     <---> %s\n",strtime(&tartime),fname);
*/
	      break;
      }
    }
    else  /* (getheader != 1) */
    {
      unsigned int bytes = (remaining > BLOCKSIZE) ? BLOCKSIZE : remaining;
	  unsigned long bwritten;

      if (outfile != INVALID_HANDLE_VALUE)
      {
          WriteFile(outfile,buffer.buffer,bytes,&bwritten,NULL);
		  if (bwritten != bytes)
          {
			  PrintMessage("Error writing %s, skipping file...", fname);
              CloseHandle(outfile);
              DeleteFile(fname);
          }
      }
      remaining -= bytes;
      if (remaining == 0)
      {
          getheader = 1;
          if (outfile != INVALID_HANDLE_VALUE)
          {
              HANDLE hFile;
              FILETIME ftm;
 
              CloseHandle(outfile);
              outfile = INVALID_HANDLE_VALUE;

              hFile = CreateFile(fname, GENERIC_READ | GENERIC_WRITE,
				     0, NULL, OPEN_EXISTING, 0, NULL);
              cnv_tar2win_time(tartime, &ftm);
              SetFileTime(hFile,&ftm,NULL,&ftm);
              CloseHandle(hFile);
          }
      }
    }
  } /* while(1) */
  
  if (gzclose(in) != Z_OK)
  {
    PrintMessage("failed gzclose");
    /* return -1; */
  }

  return 0;
}

